/**
 * Copyright (c) 2016-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *   Abel Hegedus, Tamas Borbas, Peter Lunk (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.ui.utils;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.io.Files;
import java.io.File;
import org.apache.log4j.Logger;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.emf.common.command.AbstractCommand;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.diffmerge.api.IDiffPolicy;
import org.eclipse.emf.diffmerge.api.IMatchPolicy;
import org.eclipse.emf.diffmerge.api.IMergePolicy;
import org.eclipse.emf.diffmerge.api.scopes.IEditableModelScope;
import org.eclipse.emf.diffmerge.diffdata.EComparison;
import org.eclipse.emf.diffmerge.diffdata.impl.EComparisonImpl;
import org.eclipse.emf.diffmerge.impl.policies.DefaultDiffPolicy;
import org.eclipse.emf.diffmerge.impl.policies.DefaultMatchPolicy;
import org.eclipse.emf.diffmerge.impl.policies.DefaultMergePolicy;
import org.eclipse.emf.diffmerge.impl.scopes.FragmentedModelScope;
import org.eclipse.emf.diffmerge.patch.api.IModelPatchEntryFilter;
import org.eclipse.emf.diffmerge.patch.api.ModelPatch;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchFilterApplier;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchReverser;
import org.eclipse.emf.diffmerge.patch.api.PatchApplication;
import org.eclipse.emf.diffmerge.patch.runtime.EMFModelPatchApplier;
import org.eclipse.emf.diffmerge.patch.runtime.identifier.BaseIndexEObjectLocator;
import org.eclipse.emf.diffmerge.patch.runtime.identifier.IdentifiedEMFObjectLocator;
import org.eclipse.emf.diffmerge.patch.runtime.modelaccess.EMFModelAccess;
import org.eclipse.emf.diffmerge.patch.runtime.modelaccess.ModelAccessProvider;
import org.eclipse.emf.diffmerge.patch.serializer.IModelPatchSerializer;
import org.eclipse.emf.diffmerge.patch.ui.preferences.ModelPatchPreferenceProvider;
import org.eclipse.emf.diffmerge.patch.ui.utils.SerializerProvider;
import org.eclipse.emf.diffmerge.ui.viewers.EMFDiffNode;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.viatra.query.runtime.base.api.BaseIndexOptions;
import org.eclipse.viatra.query.runtime.base.api.IndexingLevel;
import org.eclipse.viatra.query.runtime.base.api.NavigationHelper;
import org.eclipse.viatra.query.runtime.base.api.ViatraBaseFactory;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.lib.StringExtensions;

@SuppressWarnings("all")
public class ModelpatchApplicationDTO {
  @Extension
  private ModelPatchFilterApplier applier = new ModelPatchFilterApplier();
  
  @Extension
  private ModelAccessProvider provider = new ModelAccessProvider();
  
  @Extension
  private SerializerProvider serializerProvider = new SerializerProvider();
  
  @Extension
  private ModelPatchPreferenceProvider _modelPatchPreferenceProvider = ModelPatchPreferenceProvider.INSTANCE;
  
  public Resource originalModel;
  
  public Resource modifiedModel;
  
  public String patchPath;
  
  public ModelPatch originalModelPatch;
  
  public ModelPatch modifiedModelPatch;
  
  public PatchApplication patchApplication;
  
  public EMFDiffNode diffNode;
  
  private Command lastCommand;
  
  private EditingDomain editingDomain;
  
  public ModelpatchApplicationDTO(final Resource originalModel) {
    Preconditions.<Resource>checkNotNull(originalModel, "Model cannot be null.");
    this.originalModel = originalModel;
  }
  
  public boolean loadPatch(final String patchPath) {
    try {
      boolean _isNullOrEmpty = StringExtensions.isNullOrEmpty(patchPath);
      boolean _not = (!_isNullOrEmpty);
      if (_not) {
        try {
          final File file = new File(patchPath);
          final IModelPatchSerializer serializerByFileExtension = this.serializerProvider.getSerializerByFileExtension(Files.getFileExtension(patchPath));
          IModelPatchSerializer _elvis = null;
          if (serializerByFileExtension != null) {
            _elvis = serializerByFileExtension;
          } else {
            IModelPatchSerializer _selectedSerializer = this.serializerProvider.getSelectedSerializer(this._modelPatchPreferenceProvider.getSerializationType());
            _elvis = _selectedSerializer;
          }
          final IModelPatchSerializer serializer = _elvis;
          this.originalModelPatch = serializer.load(file);
          this.modifiedModelPatch = this.originalModelPatch;
        } catch (final Throwable _t) {
          if (_t instanceof Exception) {
            final Exception ex = (Exception)_t;
            this.originalModelPatch = null;
            this.patchPath = patchPath;
            throw ex;
          } else {
            throw Exceptions.sneakyThrow(_t);
          }
        }
      } else {
        this.originalModelPatch = null;
        return false;
      }
      this.patchPath = patchPath;
      return true;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public ModelPatch modifyPatch(final boolean isReverse, final IModelPatchEntryFilter... filters) {
    try {
      if (isReverse) {
        this.modifiedModelPatch = ModelPatchReverser.INSTANCE.reverse(this.originalModelPatch);
      } else {
        this.modifiedModelPatch = this.originalModelPatch;
      }
      if ((filters != null)) {
        for (final IModelPatchEntryFilter filter : filters) {
          if ((filter != null)) {
            this.modifiedModelPatch = this.applier.applyFilter(this.modifiedModelPatch, filter);
          }
        }
      }
      return this.modifiedModelPatch;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public PatchApplication applyPatch() {
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("Resolved Model");
    this.modifiedModel = this.createCopy(this.originalModel, _builder.toString());
    this.patchApplication = this.applyPatch(this.modifiedModel);
    return this.patchApplication;
  }
  
  public void saveModel() {
    abstract class __ModelpatchApplicationDTO_1 extends AbstractCommand {
      ModelPatch patch;
      
      ModelPatch reversePatch;
      
      Resource model;
      
      abstract void executePatchApplication(final Resource resource, final ModelPatch patch);
    }
    
    try {
      final __ModelpatchApplicationDTO_1 patchCommand = new __ModelpatchApplicationDTO_1() {
        @Override
        public void execute() {
          this.executePatchApplication(this.model, this.patch);
        }
        
        @Override
        public void redo() {
          this.executePatchApplication(this.model, this.patch);
        }
        
        @Override
        public void undo() {
          this.executePatchApplication(this.model, this.reversePatch);
        }
        
        @Override
        protected boolean prepare() {
          return true;
        }
        
        void executePatchApplication(final Resource resource, final ModelPatch patch) {
          final EMFModelPatchApplier patchApplier = ModelpatchApplicationDTO.this.preparePatchApplier(resource.getResourceSet());
          ModelpatchApplicationDTO.this.patchApplication = patchApplier.apply(patch, resource.getResourceSet());
        }
      };
      patchCommand.model = this.originalModel;
      patchCommand.patch = this.modifiedModelPatch;
      patchCommand.reversePatch = ModelPatchReverser.INSTANCE.reverse(this.modifiedModelPatch);
      AdapterFactoryEditingDomain.getEditingDomainFor(this.originalModel.getContents().get(0)).getCommandStack().execute(patchCommand);
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public EMFDiffNode prepareDiffNode() {
    final IEditableModelScope targetScope = new FragmentedModelScope(this.modifiedModel, true);
    this.editingDomain = AdapterFactoryEditingDomain.getEditingDomainFor(this.originalModel.getContents().get(0));
    this.lastCommand = this.editingDomain.getCommandStack().getUndoCommand();
    final IEditableModelScope referenceScope = new FragmentedModelScope(this.originalModel, false);
    final IProgressMonitor aProgressMonitor = new NullProgressMonitor();
    final IMergePolicy aMergePolicy = new DefaultMergePolicy();
    final IDiffPolicy aDiffPolicy = new DefaultDiffPolicy();
    final IMatchPolicy aMatchPolicy = new DefaultMatchPolicy();
    final EComparisonImpl comparison = new EComparisonImpl(targetScope, referenceScope);
    comparison.compute(aMatchPolicy, aDiffPolicy, aMergePolicy, aProgressMonitor);
    EMFDiffNode _eMFDiffNode = new EMFDiffNode(((EComparison) comparison), this.editingDomain, false, true);
    final Procedure1<EMFDiffNode> _function = (EMFDiffNode it) -> {
      it.setEditable(false, true);
    };
    EMFDiffNode _doubleArrow = ObjectExtensions.<EMFDiffNode>operator_doubleArrow(_eMFDiffNode, _function);
    this.diffNode = _doubleArrow;
    return this.diffNode;
  }
  
  public void undoEDMModifications() {
    if ((this.lastCommand == null)) {
      while (((!Objects.equal(this.lastCommand, this.editingDomain.getCommandStack().getUndoCommand())) && this.editingDomain.getCommandStack().canUndo())) {
        this.editingDomain.getCommandStack().undo();
      }
    } else {
      while (this.editingDomain.getCommandStack().canUndo()) {
        this.editingDomain.getCommandStack().undo();
      }
    }
  }
  
  private PatchApplication applyPatch(final Resource resource) {
    final EMFModelPatchApplier patchApplier = this.preparePatchApplier(resource.getResourceSet());
    return patchApplier.apply(this.modifiedModelPatch, resource.getResourceSet());
  }
  
  public EMFModelPatchApplier preparePatchApplier(final ResourceSet resourceSet) {
    try {
      EMFModelPatchApplier _xblockexpression = null;
      {
        final EMFModelAccess access = this.provider.getSelectedModelAccess(this._modelPatchPreferenceProvider.getModelAccessType(), resourceSet);
        final EMFModelPatchApplier patchApplier = new EMFModelPatchApplier(access);
        ViatraBaseFactory _instance = ViatraBaseFactory.getInstance();
        BaseIndexOptions _baseIndexOptions = new BaseIndexOptions(false, IndexingLevel.FULL);
        final NavigationHelper navigationHelper = _instance.createNavigationHelper(resourceSet, _baseIndexOptions, Logger.getLogger(this.getClass()));
        IdentifiedEMFObjectLocator _locator = patchApplier.getLocator();
        BaseIndexEObjectLocator _baseIndexEObjectLocator = new BaseIndexEObjectLocator(navigationHelper);
        _locator.setEObjectLocator(_baseIndexEObjectLocator);
        _xblockexpression = patchApplier;
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private Resource createCopy(final Resource copiable, final String copyPath) {
    final ResourceSetImpl copyOfOriginalRS = new ResourceSetImpl();
    final Resource copyRes = copyOfOriginalRS.createResource(URI.createURI(copyPath));
    return this.copyTo(copiable, copyRes);
  }
  
  private Resource copyTo(final Resource copiable, final Resource target) {
    final EcoreUtil.Copier copier = new EcoreUtil.Copier();
    if (((target.getContents() != null) && (target.getContents().size() > 0))) {
      target.getContents().clear();
    }
    target.getContents().addAll(copier.<EObject>copyAll(copiable.getContents()));
    copier.copyReferences();
    return target;
  }
}
